from aiostream.test_utils import (
    add_resource,
    assert_run,
    assert_cleanup,
    event_loop_policy,
)

__all__ = [
    "add_resource",
    "assert_run",
    "assert_cleanup",
    "event_loop_policy",
]



import pytest

from aiostream import stream, pipe
from aiostream.core import Stream


@pytest.mark.asyncio
async def test_concatmap(assert_run, assert_cleanup):
    def target1(x: int, *_) -> Stream[int]:
        return stream.range(x, x + 2, interval=5)

    def target2(x: int, *_) -> Stream[int]:
        return stream.range(x, x + 4, interval=1)

    def target3(x: int, *_) -> Stream[int]:
        return stream.range(0, 3, interval=1) if x else stream.throw(ZeroDivisionError)

    # Concurrent run
    with assert_cleanup() as loop:
        xs = stream.range(0, 6, 2, interval=1)
        ys = xs | pipe.concatmap(target1)
        await assert_run(ys, [0, 1, 2, 3, 4, 5])
        assert loop.steps == [1, 1, 3, 5, 5]

    # Sequential run
    with assert_cleanup() as loop:
        xs = stream.range(0, 6, 2, interval=1)
        ys = xs | pipe.concatmap(target1, task_limit=1)
        await assert_run(ys, [0, 1, 2, 3, 4, 5])
        assert loop.steps == [5, 1, 5, 1, 5]

    # Limited run
    with assert_cleanup() as loop:
        xs = stream.range(0, 6, 2, interval=1)
        ys = xs | pipe.concatmap(target1, task_limit=2)
        await assert_run(ys, [0, 1, 2, 3, 4, 5])
        assert loop.steps == [1, 4, 1, 4, 5]

    # Make sure item arrive as soon as possible
    with assert_cleanup() as loop:
        xs = stream.just(2)
        ys = xs | pipe.concatmap(target2)
        zs = ys | pipe.timeout(2)  # Sould NOT raise
        await assert_run(zs, [2, 3, 4, 5])
        assert loop.steps == [1, 1, 1]

    # An exception might get discarded if the result can be produced before the
    # processing of the exception is required
    with assert_cleanup() as loop:
        xs = stream.iterate([True, False])
        ys = xs | pipe.concatmap(target3)
        zs = ys | pipe.take(3)
        await assert_run(zs, [0, 1, 2])
        assert loop.steps == [1, 1]


@pytest.mark.asyncio
async def test_flatmap(assert_run, assert_cleanup):
    def target1(x: int, *_) -> Stream[int]:
        return stream.range(x, x + 2, interval=5)

    # Concurrent run
    with assert_cleanup() as loop:
        xs = stream.range(0, 6, 2, interval=1)
        ys = xs | pipe.flatmap(target1)
        await assert_run(ys, [0, 2, 4, 1, 3, 5])
        assert loop.steps == [1, 1, 3, 1, 1]

    # Sequential run
    with assert_cleanup() as loop:
        xs = stream.range(0, 6, 2, interval=1)
        ys = xs | pipe.flatmap(target1, task_limit=1)
        await assert_run(ys, [0, 1, 2, 3, 4, 5])
        assert loop.steps == [5, 1, 5, 1, 5]

    # Limited run
    with assert_cleanup() as loop:
        xs = stream.range(0, 6, 2, interval=1)
        ys = xs | pipe.flatmap(target1, task_limit=2)
        await assert_run(ys, [0, 2, 1, 3, 4, 5])
        assert loop.steps == [1, 4, 1, 5]


@pytest.mark.asyncio
async def test_switchmap(assert_run, assert_cleanup):
    def target1(x: int, *_) -> Stream[int]:
        return stream.range(x, x + 5, interval=1)

    def target2(x: int, *_) -> Stream[int]:
        return stream.range(x, x + 2, interval=2)

    with assert_cleanup() as loop:
        xs = stream.range(0, 30, 10, interval=3)
        ys = xs | pipe.switchmap(target1)
        await assert_run(ys, [0, 1, 2, 10, 11, 12, 20, 21, 22, 23, 24])
        assert loop.steps == [1, 1, 1, 1, 1, 1, 1, 1, 1, 1]

    # Test cleanup procedure
    with assert_cleanup() as loop:
        xs = stream.range(0, 5, interval=1)
        ys = xs | pipe.switchmap(target2)
        await assert_run(ys[:3], [0, 1, 2])
        assert loop.steps == [1, 1]



from typing import Awaitable
import pytest
import asyncio

from aiostream import stream, pipe, async_, await_
from aiostream.test_utils import add_resource


@pytest.mark.asyncio
async def test_chain(assert_run, assert_cleanup):
    with assert_cleanup():
        xs = stream.range(5) + stream.range(5, 10)
        await assert_run(xs, list(range(10)))

    with assert_cleanup():
        xs = stream.range(10, 15) | add_resource.pipe(1)
        xs += stream.range(15, 20) | add_resource.pipe(1)
        await assert_run(xs, list(range(10, 20)))

    # Empty chain (issue #95)
    xs = stream.chain()
    await assert_run(xs, [])


@pytest.mark.asyncio
async def test_zip(assert_run):
    xs = stream.range(5) | add_resource.pipe(1.0)
    ys = xs | pipe.zip(xs, xs)
    expected = [(x,) * 3 for x in range(5)]
    await assert_run(ys, expected)

    # Exceptions from iterables are propagated
    xs = stream.zip(stream.range(2), stream.throw(AttributeError))
    with pytest.raises(AttributeError):
        await xs

    # Empty zip (issue #95)
    xs = stream.zip()
    await assert_run(xs, [])

    # Strict mode (issue #118): Iterable length mismatch raises
    xs = stream.zip(stream.range(2), stream.range(1), strict=True)
    with pytest.raises(ValueError):
        await xs

    # Strict mode (issue #118): No raise for matching-length iterables
    xs = stream.zip(stream.range(2), stream.range(2), strict=True)
    await assert_run(xs, [(0, 0), (1, 1)])

    # Strict mode (issue #118): Exceptions from iterables are propagated
    xs = stream.zip(stream.range(2), stream.throw(AttributeError), strict=True)
    with pytest.raises(AttributeError):
        await xs

    # Strict mode (issue #118): Non-strict mode works as before
    xs = stream.zip(stream.range(2), stream.range(1))
    await assert_run(xs, [(0, 0)])

    # Strict mode (issue #118): In particular, we stop immediately if any
    # one iterable is exhausted, not waiting for the others
    slow_iterable_continued_after_sleep = asyncio.Event()

    async def fast_iterable():
        yield 0
        await asyncio.sleep(1)

    async def slow_iterable():
        yield 0
        await asyncio.sleep(2)
        slow_iterable_continued_after_sleep.set()

    xs = stream.zip(fast_iterable(), slow_iterable())
    await assert_run(xs, [(0, 0)])
    assert not slow_iterable_continued_after_sleep.is_set()


@pytest.mark.asyncio
async def test_map(assert_run, assert_cleanup):
    def square_target(arg: int, *_) -> int:
        return arg**2

    def sum_target(a: int, b: int, *_) -> int:
        return a + b

    async def sleep_only(arg: float, *_) -> None:
        return await asyncio.sleep(arg)

    async def sleep_and_result(arg: float, result: int, *_) -> int:
        return await asyncio.sleep(arg, result)

    def not_a_coro_function(arg: int, *_) -> Awaitable[int]:
        return asyncio.sleep(arg, arg)

    # Synchronous/simple
    with assert_cleanup():
        xs = stream.range(5) | pipe.map(square_target)
        expected = [x**2 for x in range(5)]
        await assert_run(xs, expected)

    # Synchronous/multiple
    with assert_cleanup():
        xs = stream.range(5)
        ys = xs | pipe.map(sum_target, xs)
        expected = [x * 2 for x in range(5)]
        await assert_run(ys, expected)

    # Asynchronous/simple/concurrent
    with assert_cleanup() as loop:
        xs = stream.range(1, 4) | pipe.map(sleep_only)
        expected = [None] * 3
        await assert_run(xs, expected)
        assert loop.steps == [1, 1, 1]

    # Asynchronous/simple/sequential
    with assert_cleanup() as loop:
        xs = stream.range(1, 4) | pipe.map(sleep_only, task_limit=1)
        expected = [None] * 3
        await assert_run(xs, expected)
        assert loop.steps == [1, 2, 3]

    # Asynchronous/multiple/concurrent
    with assert_cleanup() as loop:
        xs = stream.range(1, 4)
        ys = xs | pipe.map(sleep_and_result, xs)
        await assert_run(ys, [1, 2, 3])
        assert loop.steps == [1, 1, 1]

    # Asynchronous/multiple/sequential
    with assert_cleanup() as loop:
        xs = stream.range(1, 4)
        ys = xs | pipe.map(sleep_and_result, xs, task_limit=1)
        await assert_run(ys, [1, 2, 3])
        assert loop.steps == [1, 2, 3]

    # As completed
    with assert_cleanup() as loop:
        xs = stream.iterate([2, 4, 1, 3, 5])
        ys = xs | pipe.map(sleep_and_result, xs, ordered=False)
        await assert_run(ys, [1, 2, 3, 4, 5])
        assert loop.steps == [1, 1, 1, 1, 1]

    # Invalid argument
    with pytest.raises(ValueError):
        await (stream.range(1, 4) | pipe.map(sleep_only, task_limit=0))

    # Break
    with assert_cleanup() as loop:
        xs = stream.count(1)
        ys = xs | pipe.map(sleep_and_result, xs, task_limit=10)
        await assert_run(ys[:3], [1, 2, 3])
        assert loop.steps == [1, 1, 1]

    # Stuck
    with assert_cleanup():
        xs = stream.count(1)
        ys = xs | pipe.map(sleep_and_result, xs, task_limit=1) | pipe.timeout(5)
        await assert_run(ys, [1, 2, 3, 4], asyncio.TimeoutError())

    # Force await
    with assert_cleanup():
        xs = stream.iterate([1, 2, 3])
        ys = xs | pipe.map(async_(not_a_coro_function))
        await assert_run(ys, [1, 2, 3])
        assert loop.steps == [1, 1, 1]

    # Map await_
    with assert_cleanup() as loop:
        xs = stream.iterate(map(lambda x: asyncio.sleep(x, x), [1, 2, 3]))
        ys = xs | pipe.map(await_)  # type: ignore
        await assert_run(ys, [1, 2, 3])
        assert loop.steps == [1, 1, 1]


@pytest.mark.asyncio
async def test_merge(assert_run, assert_cleanup):
    with assert_cleanup() as loop:
        xs = stream.range(1, 5, 2, interval=2) | pipe.delay(1)
        ys = stream.range(0, 5, 2, interval=2) | pipe.merge(xs)
        await assert_run(ys, [0, 1, 2, 3, 4])
        assert loop.steps == [1, 1, 1, 1]

    with assert_cleanup():
        xs = stream.range(1, 5, 2, interval=2) | pipe.delay(1)
        ys = stream.range(0, 5, 2, interval=2) | pipe.merge(xs)
        await assert_run(ys[:3], [0, 1, 2])
        assert loop.steps == [1, 1]

    with assert_cleanup():
        xs = stream.just(1) + stream.never()
        ys = xs | pipe.merge(xs) | pipe.timeout(1)
        await assert_run(ys, [1, 1], asyncio.TimeoutError())
        assert loop.steps == [1]

    # Reproduce issue #65
    with assert_cleanup():
        xs = stream.iterate([1, 2])
        ys = stream.iterate([3, 4])
        zs = stream.merge(xs, ys) | pipe.take(3)
        await assert_run(zs, [1, 2, 3])

    with assert_cleanup():
        xs = stream.iterate([1, 2, 3])
        ys = stream.throw(ZeroDivisionError())
        zs = stream.merge(xs, ys) | pipe.delay(1) | pipe.take(3)
        await assert_run(zs, [1, 2, 3])

    # Silencing of a CancelledError

    async def agen1():
        if False:
            yield
        try:
            await asyncio.sleep(2)
        except asyncio.CancelledError:
            return

    async def agen2():
        yield 1

    with assert_cleanup():
        xs = stream.merge(agen1(), agen2()) | pipe.delay(1) | pipe.take(1)
        await assert_run(xs, [1])

    # Empty merge (issue #95)
    xs = stream.merge()
    await assert_run(xs, [])


@pytest.mark.asyncio
async def test_ziplatest(assert_run, assert_cleanup):
    with assert_cleanup() as loop:
        xs = stream.range(0, 5, 2, interval=2)
        ys = stream.range(1, 5, 2, interval=2) | pipe.delay(1)
        zs = stream.ziplatest(xs, ys, default="▲")
        await assert_run(zs, [(0, "▲"), (0, 1), (2, 1), (2, 3), (4, 3)])
        assert loop.steps == [1, 1, 1, 1]

    with assert_cleanup() as loop:
        xs = stream.range(0, 5, 2, interval=2)
        ys = stream.range(1, 5, 2, interval=2) | pipe.delay(1)
        zs = stream.ziplatest(xs, ys, partial=False)
        await assert_run(zs, [(0, 1), (2, 1), (2, 3), (4, 3)])
        assert loop.steps == [1, 1, 1, 1]

    # Empty ziplatest (issue #95)
    xs = stream.ziplatest()
    await assert_run(xs, [])



import pytest
import asyncio
import operator

from aiostream import stream, pipe
from aiostream.test_utils import add_resource


@pytest.mark.asyncio
async def test_accumulate(assert_run, assert_cleanup):
    with assert_cleanup():
        xs = stream.range(5) | add_resource.pipe(1) | pipe.accumulate()
        await assert_run(xs, [0, 1, 3, 6, 10])

    with assert_cleanup():
        xs = (
            stream.range(2, 4)
            | add_resource.pipe(1)
            | pipe.accumulate(func=operator.mul, initializer=2)
        )
        await assert_run(xs, [2, 4, 12])

    with assert_cleanup():
        xs = stream.range(0) | add_resource.pipe(1) | pipe.accumulate()
        await assert_run(xs, [])

    async def sleepmax(x, y):
        return await asyncio.sleep(1, result=max(x, y))

    with assert_cleanup() as loop:
        xs = stream.range(3) | add_resource.pipe(1) | pipe.accumulate(sleepmax)
        await assert_run(xs, [0, 1, 2])
        assert loop.steps == [1] * 3


@pytest.mark.asyncio
async def test_reduce(assert_run, assert_cleanup):
    with assert_cleanup():
        xs = stream.range(5) | add_resource.pipe(1) | pipe.reduce(min)
        await assert_run(xs, [0])

    with assert_cleanup():
        xs = stream.range(5) | add_resource.pipe(1) | pipe.reduce(max)
        await assert_run(xs, [4])

    with assert_cleanup():
        xs = stream.range(0) | add_resource.pipe(1) | pipe.reduce(max)
        await assert_run(xs, [], IndexError("Index out of range"))


@pytest.mark.asyncio
async def test_list(assert_run, assert_cleanup):
    with assert_cleanup():
        xs = stream.range(3) | add_resource.pipe(1) | pipe.list()
        # The same list object is yielded at each step
        await assert_run(xs, [[0, 1, 2], [0, 1, 2], [0, 1, 2], [0, 1, 2]])

    with assert_cleanup():
        xs = stream.range(0) | add_resource.pipe(1) | pipe.list()
        await assert_run(xs, [[]])



import pytest
import asyncio

from aiostream.aiter_utils import AsyncIteratorContext, aitercontext, anext


# Some async iterators for testing


async def agen():
    for x in range(5):
        await asyncio.sleep(1)
        yield x


async def silence_agen():
    while True:
        try:
            yield 1
        except BaseException:
            return


async def reraise_agen():
    while True:
        try:
            yield 1
        except BaseException as exc:
            raise RuntimeError from exc


async def stuck_agen():
    try:
        yield 1
    except BaseException:
        yield 2


class not_an_agen(list):
    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            return self.pop()
        except IndexError:
            raise StopAsyncIteration


# Tests


@pytest.mark.asyncio
async def test_simple_aitercontext(assert_cleanup):
    with assert_cleanup() as loop:
        async with aitercontext(agen()) as safe_gen:
            # Cannot enter twice
            with pytest.raises(RuntimeError):
                async with safe_gen:
                    pass

            it = iter(range(5))
            async for item in safe_gen:
                assert item == next(it)
        assert loop.steps == [1] * 5

    # Exiting is idempotent
    await safe_gen.__aexit__(None, None, None)
    await safe_gen.aclose()

    with pytest.raises(RuntimeError):
        await anext(safe_gen)

    with pytest.raises(RuntimeError):
        async with safe_gen:
            pass

    with pytest.raises(RuntimeError):
        await safe_gen.athrow(ValueError())


@pytest.mark.asyncio
async def test_athrow_in_aitercontext():
    async with aitercontext(agen()) as safe_gen:
        assert await safe_gen.__anext__() == 0
        with pytest.raises(ZeroDivisionError):
            await safe_gen.athrow(ZeroDivisionError())
        async for _ in safe_gen:
            assert False  # No more items


@pytest.mark.asyncio
async def test_aitercontext_wrong_usage():
    safe_gen = aitercontext(agen())
    with pytest.warns(UserWarning):
        await anext(safe_gen)

    with pytest.raises(TypeError):
        AsyncIteratorContext(None)  # type: ignore

    with pytest.raises(TypeError):
        AsyncIteratorContext(safe_gen)


@pytest.mark.asyncio
async def test_raise_in_aitercontext():
    with pytest.raises(ZeroDivisionError):
        async with aitercontext(agen()) as safe_gen:
            async for _ in safe_gen:
                raise ZeroDivisionError

    with pytest.raises(ZeroDivisionError):
        async with aitercontext(agen()) as safe_gen:
            async for _ in safe_gen:
                pass
            raise ZeroDivisionError

    with pytest.raises(GeneratorExit):
        async with aitercontext(agen()) as safe_gen:
            async for _ in safe_gen:
                raise GeneratorExit

    with pytest.raises(GeneratorExit):
        async with aitercontext(agen()) as safe_gen:
            async for _ in safe_gen:
                pass
            raise GeneratorExit


@pytest.mark.asyncio
async def test_silence_exception_in_aitercontext():
    async with aitercontext(silence_agen()) as safe_gen:
        async for item in safe_gen:
            assert item == 1
            raise ZeroDivisionError

    # Silencing a generator exit is forbidden
    with pytest.raises(GeneratorExit):
        async with aitercontext(silence_agen()) as safe_gen:
            async for _ in safe_gen:
                raise GeneratorExit


@pytest.mark.asyncio
async def test_reraise_exception_in_aitercontext():
    with pytest.raises(RuntimeError) as info:
        async with aitercontext(reraise_agen()) as safe_gen:
            async for item in safe_gen:
                assert item == 1
                raise ZeroDivisionError
    assert type(info.value.__cause__) is ZeroDivisionError

    with pytest.raises(RuntimeError) as info:
        async with aitercontext(reraise_agen()) as safe_gen:
            async for item in safe_gen:
                assert item == 1
                raise GeneratorExit
    assert type(info.value.__cause__) is GeneratorExit


@pytest.mark.asyncio
async def test_stuck_in_aitercontext():
    with pytest.raises(RuntimeError) as info:
        async with aitercontext(stuck_agen()) as safe_gen:
            async for item in safe_gen:
                assert item == 1
                raise ZeroDivisionError
    assert "didn't stop after athrow" in str(info.value)

    with pytest.raises(RuntimeError) as info:
        async with aitercontext(stuck_agen()) as safe_gen:
            async for item in safe_gen:
                assert item == 1
                raise GeneratorExit
    # GeneratorExit relies on aclose, not athrow
    # The message is a bit different
    assert "async generator ignored GeneratorExit" in str(info.value)

    # Check athrow behavior
    async with aitercontext(stuck_agen()) as safe_gen:
        assert await safe_gen.__anext__() == 1
        assert await safe_gen.athrow(ZeroDivisionError()) == 2


@pytest.mark.asyncio
async def test_not_an_agen_in_aitercontext():
    async with aitercontext(not_an_agen([1])) as safe_gen:
        async for item in safe_gen:
            assert item == 1

    with pytest.raises(ZeroDivisionError):
        async with aitercontext(not_an_agen([1])) as safe_gen:
            async for item in safe_gen:
                raise ZeroDivisionError


@pytest.mark.asyncio
async def test_aitercontext_idempotent():
    async with aitercontext(aitercontext(agen())) as safe_gen:
        async for item in safe_gen:
            assert item == 0
            break



import asyncio

import pytest
from aiostream.manager import TaskGroup


async def task1():
    await asyncio.sleep(1)
    return 1


async def task2():
    await asyncio.sleep(0)
    return 2


async def task3():
    await asyncio.sleep(0)
    raise ValueError(3)


async def task4():
    await asyncio.sleep(1)
    raise ValueError(4)


async def task5():
    try:
        await asyncio.sleep(1)
    finally:
        raise ValueError(5)


@pytest.mark.asyncio
async def test_task_group_cleanup():
    async with TaskGroup() as group:
        t1 = group.create_task(task1())
        t2 = group.create_task(task2())
        t3 = group.create_task(task3())
        t4 = group.create_task(task4())
        t5 = group.create_task(task5())
        # Cancel task1
        t1.cancel()
        with pytest.raises(asyncio.CancelledError):
            await t1
        assert t1.cancelled()
        # Join task2
        assert await t2 == 2
        # Task 3 is finished
        assert t3.done()
        # Task 4 and 5 are not finished
        assert not t4.done()
        assert not t5.done()



import inspect
import pytest

from aiostream.core import pipable_operator, sources_operator
from aiostream.test_utils import add_resource
from aiostream import stream, streamcontext, operator


@pytest.mark.asyncio
async def test_streamcontext(assert_cleanup):
    with assert_cleanup() as loop:
        xs = stream.range(3) | add_resource.pipe(1)
        async with streamcontext(xs) as streamer:
            it = iter(range(3))
            async for item in streamer:
                assert item == next(it)
        assert loop.steps == [1]

    with assert_cleanup() as loop:
        xs = stream.range(5) | add_resource.pipe(1)
        async with xs.stream() as streamer:
            it = iter(range(5))
            async for item in streamer:
                assert item == next(it)
        assert loop.steps == [1]


@pytest.mark.parametrize(
    "operator_param", [operator, pipable_operator, sources_operator]
)
def test_operator_from_method(operator_param):
    with pytest.raises(ValueError):

        class A:
            @operator_param
            async def method(self, arg):
                yield 1

    with pytest.raises(ValueError):

        class B:
            @operator_param
            async def method(cls, arg):
                yield 1

    with pytest.raises(ValueError):

        class C:
            @operator_param
            @classmethod
            async def method(cls, arg):
                yield 1


@pytest.mark.asyncio
async def test_error_on_sync_iteration():
    xs = stream.range(3)

    # Stream raises a TypeError
    with pytest.raises(TypeError):
        for x in xs:
            assert False

    # Streamer raises a TypeError
    async with xs.stream() as streamer:
        with pytest.raises(TypeError):
            for x in streamer:
                assert False


@pytest.mark.asyncio
async def test_error_on_entering_a_stream():
    xs = stream.range(3)

    # Stream raises a TypeError
    with pytest.raises(TypeError) as ctx:
        async with xs:  # type: ignore
            assert False

    assert "Use the `stream` method" in str(ctx.value)


def test_compatibility():
    @operator
    async def test1():
        yield 1

    with pytest.raises(AttributeError):
        test1.pipe  # type: ignore

    match = "The `pipable` argument is deprecated."
    with pytest.warns(DeprecationWarning, match=match):

        @operator()
        async def test2():
            yield 1

    with pytest.raises(AttributeError):
        test2.pipe  # type: ignore

    with pytest.warns(DeprecationWarning, match=match):

        @operator(pipable=False)
        async def test3():
            yield 1

    with pytest.raises(AttributeError):
        test3.pipe  # type: ignore

    with pytest.warns(DeprecationWarning, match=match):

        @operator(pipable=True)
        async def test4(source):
            yield 1

    test4.pipe  # type: ignore

    with pytest.warns(DeprecationWarning, match=match):

        async def test5(source):
            yield 1

        test5_operator = operator(test5, pipable=True)
        test5_operator.pipe  # type: ignore


def test_pipable_operator_with_variadic_args():

    with pytest.raises(ValueError):

        @pipable_operator
        async def test1(*args):
            yield 1


def test_sources_operator_with_postional_args():
    with pytest.raises(ValueError):

        @sources_operator
        async def test1(source):
            yield 1


def test_introspection_for_operator():
    # Extract original information
    original = stream.range.original  # type: ignore
    original_doc = original.__doc__
    assert original_doc is not None
    assert original_doc.splitlines()[0] == "Generate a given range of numbers."
    assert (
        str(inspect.signature(original))
        == "(*args: 'int', interval: 'float' = 0.0) -> 'AsyncIterator[int]'"
    )

    # Check the stream operator
    assert str(stream.range) == repr(stream.range) == "aiostream.stream.create.range"
    assert stream.range.__module__ == "aiostream.stream.create"
    assert stream.range.__doc__ == original_doc

    # Check the raw method
    assert stream.range.raw.__qualname__ == "range.raw"
    assert stream.range.raw.__module__ == "aiostream.stream.create"
    assert stream.range.raw.__doc__ == original_doc
    assert (
        str(inspect.signature(stream.range.raw))
        == "(*args: 'int', interval: 'float' = 0.0) -> 'AsyncIterator[int]'"
    )

    # Check the __call__ method
    assert stream.range.__call__.__qualname__ == "range.__call__"
    assert stream.range.__call__.__module__ == "aiostream.stream.create"
    assert stream.range.__call__.__doc__ == original_doc
    assert (
        str(inspect.signature(stream.range.__call__))
        == "(*args: 'int', interval: 'float' = 0.0) -> 'Stream[int]'"
    )


def test_introspection_for_pipable_operator():
    # Extract original information
    original = stream.take.original  # type: ignore
    original_doc = original.__doc__
    assert original_doc is not None
    assert (
        original_doc.splitlines()[0]
        == "Forward the first ``n`` elements from an asynchronous sequence."
    )
    assert (
        str(inspect.signature(original))
        == "(source: 'AsyncIterable[T]', n: 'int') -> 'AsyncIterator[T]'"
    )

    # Check the stream operator
    assert str(stream.take) == repr(stream.take) == "aiostream.stream.select.take"
    assert stream.take.__module__ == "aiostream.stream.select"
    assert stream.take.__doc__ == original_doc

    # Check the raw method
    assert stream.take.raw.__qualname__ == "take.raw"
    assert stream.take.raw.__module__ == "aiostream.stream.select"
    assert stream.take.raw.__doc__ == original_doc
    assert (
        str(inspect.signature(stream.take.raw))
        == "(source: 'AsyncIterable[T]', n: 'int') -> 'AsyncIterator[T]'"
    )

    # Check the __call__ method
    assert stream.take.__call__.__qualname__ == "take.__call__"
    assert stream.take.__call__.__module__ == "aiostream.stream.select"
    assert stream.take.__call__.__doc__ == original_doc
    assert (
        str(inspect.signature(stream.take.__call__))
        == "(source: 'AsyncIterable[T]', n: 'int') -> 'Stream[T]'"
    )

    # Check the pipe method
    assert stream.take.pipe.__qualname__ == "take.pipe"
    assert stream.take.pipe.__module__ == "aiostream.stream.select"
    assert (
        stream.take.pipe.__doc__
        == 'Piped version of the "take" stream operator.\n\n    ' + original_doc
    )
    assert (
        str(inspect.signature(stream.take.pipe))
        == "(n: 'int') -> 'Callable[[AsyncIterable[X]], Stream[T]]'"
    )


def test_introspection_for_sources_operator():
    # Extract original information
    original = stream.zip.original  # type: ignore
    original_doc = original.__doc__
    assert original_doc is not None
    assert (
        original_doc.splitlines()[0]
        == "Combine and forward the elements of several asynchronous sequences."
    )
    assert (
        str(inspect.signature(original))
        == "(*sources: 'AsyncIterable[T]', strict: 'bool' = False) -> 'AsyncIterator[tuple[T, ...]]'"
    )

    # Check the stream operator
    assert str(stream.zip) == repr(stream.zip) == "aiostream.stream.combine.zip"
    assert stream.zip.__module__ == "aiostream.stream.combine"
    assert stream.zip.__doc__ == original_doc

    # Check the raw method
    assert stream.zip.raw.__qualname__ == "zip.raw"
    assert stream.zip.raw.__module__ == "aiostream.stream.combine"
    assert stream.zip.raw.__doc__ == original_doc
    assert (
        str(inspect.signature(stream.zip.raw))
        == "(*sources: 'AsyncIterable[T]', strict: 'bool' = False) -> 'AsyncIterator[tuple[T, ...]]'"
    )

    # Check the __call__ method
    assert stream.zip.__call__.__qualname__ == "zip.__call__"
    assert stream.zip.__call__.__module__ == "aiostream.stream.combine"
    assert stream.zip.__call__.__doc__ == original_doc
    assert (
        str(inspect.signature(stream.zip.__call__))
        == "(*sources: 'AsyncIterable[T]', strict: 'bool' = False) -> 'Stream[tuple[T, ...]]'"
    )

    # Check the pipe method
    assert stream.zip.pipe.__qualname__ == "zip.pipe"
    assert stream.zip.pipe.__module__ == "aiostream.stream.combine"
    assert (
        stream.zip.pipe.__doc__
        == 'Piped version of the "zip" stream operator.\n\n    ' + original_doc
    )
    assert (
        str(inspect.signature(stream.zip.pipe))
        == "(*sources: 'AsyncIterable[T]', strict: 'bool' = False) -> 'Callable[[AsyncIterable[Any]], Stream[tuple[T, ...]]]'"
    )



import pytest
import asyncio

from aiostream import stream, pipe


@pytest.mark.asyncio
async def test_timeout(assert_run, assert_cleanup):
    with assert_cleanup() as loop:
        xs = stream.range(3) | pipe.timeout(5)
        await assert_run(xs, [0, 1, 2])
        assert loop.steps == []

    with assert_cleanup():
        xs = stream.range(3) + stream.never()
        ys = xs | pipe.timeout(1)
        await assert_run(ys, [0, 1, 2], asyncio.TimeoutError())
        assert loop.steps == [1]



import pytest
import asyncio

from aiostream import stream, pipe
from aiostream.test_utils import add_resource


@pytest.mark.asyncio
async def test_starmap(assert_run, assert_cleanup):
    def target(a: int, b: int, *_) -> int:
        return a + b

    with assert_cleanup():
        xs = stream.range(5)
        ys = stream.range(5)
        zs = xs | pipe.zip(ys) | pipe.starmap(target)
        expected = [x * 2 for x in range(5)]
        await assert_run(zs, expected)

    async def async_target(arg: float, result: int, *_) -> int:
        return await asyncio.sleep(arg, result)

    with assert_cleanup() as loop:
        xs = stream.range(1, 4)
        ys = stream.range(1, 4)
        zs = xs | pipe.zip(ys) | pipe.starmap(async_target)
        await assert_run(zs, [1, 2, 3])
        assert loop.steps == [1, 1, 1]

    with assert_cleanup() as loop:
        xs = stream.range(1, 4)
        ys = stream.range(1, 4)
        zs = xs | pipe.zip(ys) | pipe.starmap(async_target, task_limit=1)
        await assert_run(zs, [1, 2, 3])
        assert loop.steps == [1, 2, 3]


@pytest.mark.asyncio
async def test_cycle(assert_run, assert_cleanup):
    with assert_cleanup():
        xs = stream.empty() | pipe.cycle() | pipe.timeout(1)
        await assert_run(xs, [], asyncio.TimeoutError())

    with assert_cleanup():
        xs = stream.empty() | add_resource.pipe(1) | pipe.cycle() | pipe.timeout(1)
        await assert_run(xs, [], asyncio.TimeoutError())

    with assert_cleanup() as loop:
        xs = stream.just(1) | add_resource.pipe(1) | pipe.cycle()
        await assert_run(xs[:5], [1] * 5)
        assert loop.steps == [1] * 5


@pytest.mark.asyncio
async def test_chunks(assert_run, assert_cleanup):
    with assert_cleanup():
        xs = stream.range(3, interval=1) | pipe.chunks(3)
        await assert_run(xs, [[0, 1, 2]])

    with assert_cleanup():
        xs = stream.range(4, interval=1) | pipe.chunks(3)
        await assert_run(xs, [[0, 1, 2], [3]])

    with assert_cleanup():
        xs = stream.range(5, interval=1) | pipe.chunks(3)
        await assert_run(xs, [[0, 1, 2], [3, 4]])

    with assert_cleanup():
        xs = stream.count(interval=1) | add_resource.pipe(1) | pipe.chunks(3)
        await assert_run(xs[:1], [[0, 1, 2]])



from aiostream import stream, pipe


def test_pipe_module():
    for name in dir(stream):
        obj = getattr(stream, name)
        pipe_method = getattr(obj, "pipe", None)
        if pipe_method is None:
            continue
        assert getattr(pipe, name) == pipe_method
